# -*- coding: utf-8 -*-
# File: imgproc.py


import numpy as np
import cv2

from ...utils.develop import log_deprecated
from .base import PhotometricAugmentor

__all__ = ['Hue', 'Brightness', 'BrightnessScale', 'Contrast', 'MeanVarianceNormalize',
           'GaussianBlur', 'Gamma', 'Clip', 'Saturation', 'Lighting', 'MinMaxNormalize']


class Hue(PhotometricAugmentor):
    """ Randomly change color hue.
    """

    def __init__(self, range=(0, 180), rgb=True):
        """
        Args:
            range(list or tuple): range from which the applied hue offset is selected
                (maximum range can be [-90,90] for both uint8 and float32)
            rgb (bool): whether input is RGB or BGR.
        """
        super(Hue, self).__init__()
        rgb = bool(rgb)
        self._init(locals())

    def _get_augment_params(self, _):
        return self._rand_range(*self.range)

    def _augment(self, img, hue):
        m = cv2.COLOR_BGR2HSV if not self.rgb else cv2.COLOR_RGB2HSV
        hsv = cv2.cvtColor(img, m)
        # https://docs.opencv.org/3.2.0/de/d25/imgproc_color_conversions.html#color_convert_rgb_hsv
        if hsv.dtype.itemsize == 1:
            # OpenCV uses 0-179 for 8-bit images
            hsv[..., 0] = (hsv[..., 0] + hue) % 180
        else:
            # OpenCV uses 0-360 for floating point images
            hsv[..., 0] = (hsv[..., 0] + 2 * hue) % 360
        m = cv2.COLOR_HSV2BGR if not self.rgb else cv2.COLOR_HSV2RGB
        img = cv2.cvtColor(hsv, m)
        return img


class Brightness(PhotometricAugmentor):
    """
    Adjust brightness by adding a random number.
    """
    def __init__(self, delta, clip=True):
        """
        Args:
            delta (float): Randomly add a value within [-delta,delta]
            clip (bool): clip results to [0,255] even when data type is not uint8.
        """
        super(Brightness, self).__init__()
        assert delta > 0
        self._init(locals())

    def _get_augment_params(self, _):
        return self._rand_range(-self.delta, self.delta)

    def _augment(self, img, v):
        old_dtype = img.dtype
        img = img.astype('float32')
        img += v
        if self.clip or old_dtype == np.uint8:
            img = np.clip(img, 0, 255)
        return img.astype(old_dtype)


class BrightnessScale(PhotometricAugmentor):
    """
    Adjust brightness by scaling by a random factor.
    """
    def __init__(self, range, clip=True):
        """
        Args:
            range (tuple): Randomly scale the image by a factor in (range[0], range[1])
            clip (bool): clip results to [0,255] even when data type is not uint8.
        """
        super(BrightnessScale, self).__init__()
        self._init(locals())

    def _get_augment_params(self, _):
        return self._rand_range(*self.range)

    def _augment(self, img, v):
        old_dtype = img.dtype
        img = img.astype('float32')
        img *= v
        if self.clip or old_dtype == np.uint8:
            img = np.clip(img, 0, 255)
        return img.astype(old_dtype)


class Contrast(PhotometricAugmentor):
    """
    Apply ``x = (x - mean) * contrast_factor + mean`` to each channel.
    """

    def __init__(self, factor_range, rgb=None, clip=True):
        """
        Args:
            factor_range (list or tuple): an interval to randomly sample the `contrast_factor`.
            rgb (bool or None): if None, use the mean per-channel.
            clip (bool): clip to [0, 255] even when data type is not uint8.
        """
        super(Contrast, self).__init__()
        self._init(locals())

    def _get_augment_params(self, _):
        return self._rand_range(*self.factor_range)

    def _augment(self, img, r):
        old_dtype = img.dtype

        if img.ndim == 3:
            if self.rgb is not None:
                m = cv2.COLOR_RGB2GRAY if self.rgb else cv2.COLOR_BGR2GRAY
                grey = cv2.cvtColor(img.astype('float32'), m)
                mean = np.mean(grey)
            else:
                mean = np.mean(img, axis=(0, 1), keepdims=True)
        else:
            mean = np.mean(img)

        img = img * r + mean * (1 - r)
        if self.clip or old_dtype == np.uint8:
            img = np.clip(img, 0, 255)
        return img.astype(old_dtype)


class MeanVarianceNormalize(PhotometricAugmentor):
    """
    Linearly scales the image to have zero mean and unit norm.
    ``x = (x - mean) / adjusted_stddev``
    where ``adjusted_stddev = max(stddev, 1.0/sqrt(num_pixels * channels))``

    This augmentor always returns float32 images.
    """

    def __init__(self, all_channel=True):
        """
        Args:
            all_channel (bool): if True, normalize all channels together. else separately.
        """
        self._init(locals())

    def _augment(self, img, _):
        img = img.astype('float32')
        if self.all_channel:
            mean = np.mean(img)
            std = np.std(img)
        else:
            mean = np.mean(img, axis=(0, 1), keepdims=True)
            std = np.std(img, axis=(0, 1), keepdims=True)
        std = np.maximum(std, 1.0 / np.sqrt(np.prod(img.shape)))
        img = (img - mean) / std
        return img


class GaussianBlur(PhotometricAugmentor):
    """ Gaussian blur the image with random window size"""

    def __init__(self, size_range=(0, 3), sigma_range=(0, 0), symmetric=True, max_size=None):
        """
        Args:
            size_range (tuple[int]): Gaussian window size would be 2 * size +
                1, where size is randomly sampled from this [low, high) range.
            sigma_range (tuple[float]): min,max of the sigma value. 0 means
                opencv's default.
            symmetric (bool): whether to use the same size & sigma for x and y.
            max_size (int): deprecated
        """
        super(GaussianBlur, self).__init__()
        if not isinstance(size_range, (list, tuple)):
            size_range = (0, size_range)
        assert isinstance(sigma_range, (list, tuple)), sigma_range
        if max_size is not None:
            log_deprecated("GaussianBlur(max_size=)", "Use size_range= instead!", "2020-09-01")
            size_range = (0, max_size)
        self._init(locals())

    def _get_augment_params(self, _):
        size_xy = self.rng.randint(self.size_range[0], self.size_range[1], size=(2,)) * 2 + 1
        sigma_xy = self._rand_range(*self.sigma_range, size=(2,))
        if self.symmetric:
            size_xy[1] = size_xy[0]
            sigma_xy[1] = sigma_xy[0]
        return tuple(size_xy), tuple(sigma_xy)

    def _augment(self, img, prm):
        size, sigma = prm
        return np.reshape(cv2.GaussianBlur(img, size, sigmaX=sigma[0], sigmaY=sigma[1],
                                           borderType=cv2.BORDER_REPLICATE), img.shape)


class Gamma(PhotometricAugmentor):
    """ Randomly adjust gamma """
    def __init__(self, range=(-0.5, 0.5)):
        """
        Args:
            range(list or tuple): gamma range
        """
        super(Gamma, self).__init__()
        self._init(locals())

    def _get_augment_params(self, _):
        return self._rand_range(*self.range)

    def _augment(self, img, gamma):
        old_dtype = img.dtype
        lut = ((np.arange(256, dtype='float32') / 255) ** (1. / (1. + gamma)) * 255).astype('uint8')
        img = np.clip(img, 0, 255).astype('uint8')
        ret = cv2.LUT(img, lut).astype(old_dtype)
        if img.ndim == 3 and ret.ndim == 2:
            ret = ret[:, :, np.newaxis]
        return ret


class Clip(PhotometricAugmentor):
    """ Clip the pixel values """

    def __init__(self, min=0, max=255):
        """
        Args:
            min, max: the clip range
        """
        self._init(locals())

    def _augment(self, img, _):
        return np.clip(img, self.min, self.max)


class Saturation(PhotometricAugmentor):
    """ Randomly adjust saturation.
        Follows the implementation in `fb.resnet.torch
        <https://github.com/facebook/fb.resnet.torch/blob/master/datasets/transforms.lua#L218>`__.
    """

    def __init__(self, alpha=0.4, rgb=True, clip=True):
        """
        Args:
            alpha(float): maximum saturation change.
            rgb (bool): whether input is RGB or BGR.
            clip (bool): clip results to [0,255] even when data type is not uint8.
        """
        super().__init__()
        rgb = bool(rgb)
        assert alpha < 1
        self._init(locals())

    def _get_augment_params(self, _):
        return 1 + self._rand_range(-self.alpha, self.alpha)

    def _augment(self, img, v):
        old_dtype = img.dtype
        m = cv2.COLOR_RGB2GRAY if self.rgb else cv2.COLOR_BGR2GRAY
        grey = cv2.cvtColor(img, m)
        ret = img * v + (grey * (1 - v))[:, :, np.newaxis]
        if self.clip or old_dtype == np.uint8:
            ret = np.clip(ret, 0, 255)
        return ret.astype(old_dtype)


class Lighting(PhotometricAugmentor):
    """ Lighting noise, as in the paper
        `ImageNet Classification with Deep Convolutional Neural Networks
        <https://papers.nips.cc/paper/4824-imagenet-classification-with-deep-convolutional-neural-networks.pdf>`_.
        The implementation follows `fb.resnet.torch
        <https://github.com/facebook/fb.resnet.torch/blob/master/datasets/transforms.lua#L184>`__.
    """

    def __init__(self, std, eigval, eigvec, clip=True):
        """
        Args:
            std (float): maximum standard deviation
            eigval: a vector of (3,). The eigenvalues of 3 channels.
            eigvec: a 3x3 matrix. Each column is one eigen vector.
            clip (bool): clip results to [0,255] even when data type is not uint8.
        """
        super(Lighting, self).__init__()
        eigval = np.asarray(eigval, dtype="float32")
        eigvec = np.asarray(eigvec, dtype="float32")
        assert eigval.shape == (3,)
        assert eigvec.shape == (3, 3)
        self._init(locals())

    def _get_augment_params(self, img):
        assert img.shape[2] == 3
        return (self.rng.randn(3) * self.std).astype("float32")

    def _augment(self, img, v):
        old_dtype = img.dtype
        v = v * self.eigval
        v = v.reshape((3, 1))
        inc = np.dot(self.eigvec, v).reshape((3,))
        img = np.add(img, inc)
        if self.clip or old_dtype == np.uint8:
            img = np.clip(img, 0, 255)
        return img.astype(old_dtype)


class MinMaxNormalize(PhotometricAugmentor):
    """
    Linearly scales the image to the range [min, max].

    This augmentor always returns float32 images.
    """
    def __init__(self, min=0, max=255, all_channel=True):
        """
        Args:
            max (float): The new maximum value
            min (float): The new minimum value
            all_channel (bool): if True, normalize all channels together. else separately.
        """
        self._init(locals())

    def _augment(self, img, _):
        img = img.astype('float32')
        if self.all_channel:
            minimum = np.min(img)
            maximum = np.max(img)
        else:
            minimum = np.min(img, axis=(0, 1), keepdims=True)
            maximum = np.max(img, axis=(0, 1), keepdims=True)
        img = (self.max - self.min) * (img - minimum) / (maximum - minimum) + self.min
        return img



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, time
from pathlib import Path
from loguru import logger
from .trainer import generate_dataloader
#  from .onnx import export_onnx, inference_from_onnx
from .ner_utils import LabeledText, show_ner_datainfo, get_ner_preds_reviews, save_ner_preds, load_ner_examples, load_ner_labeled_examples
from .ner_utils import to_train_poplar, to_reviews_poplar, to_sampling_poplar, to_poplar, ner_evaluate
from .glue_utils import show_glue_datainfo, load_glue_examples, save_glue_preds
from .spo_utils import show_spo_datainfo, load_spo_examples, save_spo_preds
from .common_args import add_common_args
from .utils import Params, CommonParams, NerParams, GlueParams, NerAppParams, GlueAppParams, SpoParams, SpoAppParams
from .utils import log_global_params, archive_local_model
from .utils import tensor_to_numpy, tensor_to_list, save_args
from .dataset_ner import NerDataset, ner_data_generator, merge_ner_datasets, mix_ner_datasets, load_ner_dataset
from .dataset_spo import SpoDataset, spo_data_generator, merge_spo_datasets, load_spo_dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from abc import ABCMeta, abstractmethod


class BaseRunner(metaclass=ABCMeta):
    """
    - run()
    - train()
    - evaluate()
    - save_checkpoint()
    """
    def __init__(self,
                 model,
                 batch_processor=None,
                 optimizer=None,
                 work_dir=None,
                 meta=None,
                 max_epochs=None,
                 max_iters=None):
        self.model = model
        self.batch_processor = batch_processor
        self.optimizer = optimizer
        self.meta = meta
        if isinstance(work_dir, str):
            self.work_dir = os.path.abspath(work_dir)
        else:
            self.work_dir = None
        if hasattr(self.model, 'module'):
            self._model_name = self.model.module.__class__.__name__
        else:
            self._model_name = self.model.__class__.__name__
        self.hooks = []
        self._epoch = 0
        self._iter = 0
        self._inner_iter = 0

        if max_epochs is not None and max_iters is not None:
            raise ValueError(
                "Only one of 'max_epochs' or 'max_iters' can be set.")
        self._max_epochs = max_epochs
        self._max_iters = max_iters

        @property
        def model_nae(self):
            return self._model_name

        @property
        def hooks(self):
            return self._hooks

        @property
        def epoch(self):
            return self._epoch

        @property
        def iter(self):
            return self._iter

        @property
        def inner_iter(self):
            return self._inner_iter

        @property
        def max_epochs(self):
            return self._max_epochs

        @property
        def max_iters(self):
            return self._max_iters

        @abstractmethod
        def train(self):
            pass

        @abstractmethod
        def valuate(self):
            pass

        @abstractmethod
        def run(self, data_loaders, workflow, **kwargs):
            pass

        @abstractmethod
        def save_checkpoint(self,
                            out_dir,
                            filename_tmpl,
                            save_optimizer=False,
                            meta=None,
                            create_symlink=True):
            pass

        def current_lr(self):
            if isinstance(self.optimizer, torch.optim.Optimizer):
                lr = [group['lr'] for group in self.optimizer.param_groups]
            elif isinstance(self.optimizer, dict):
                lr = dict()
                for name, optim in self.optimizer.items():
                    lr[name] = [group['lr'] for group in optim.param_groups]
            else:
                raise RuntimeError(
                    f"Optimizer type {type(self.optimizer)} is not in 'torch.optim.Optimizer' or 'dict'."
                )
            return lr

        def load_checkpoint(self, filename, map_location='cpu', strict=False):
            pass



# -*- coding: utf-8 -*-
# File: timer.py


import atexit
from collections import defaultdict
from contextlib import contextmanager
from time import perf_counter as timer  # noqa

from . import logger
from .stats import StatCounter


__all__ = ['timed_operation', 'IterSpeedCounter', 'Timer']


@contextmanager
def timed_operation(msg, log_start=False):
    """
    Surround a context with a timer.

    Args:
        msg(str): the log to print.
        log_start(bool): whether to print also at the beginning.

    Example:
        .. code-block:: python

            with timed_operation('Good Stuff'):
                time.sleep(1)

        Will print:

        .. code-block:: python

            Good stuff finished, time:1sec.
    """
    assert len(msg)
    if log_start:
        logger.info('Start {} ...'.format(msg))
    start = timer()
    yield
    msg = msg[0].upper() + msg[1:]
    logger.info('{} finished, time:{:.4f} sec.'.format(
        msg, timer() - start))


_TOTAL_TIMER_DATA = defaultdict(StatCounter)


@contextmanager
def total_timer(msg):
    """ A context which add the time spent inside to the global TotalTimer. """
    start = timer()
    yield
    t = timer() - start
    _TOTAL_TIMER_DATA[msg].feed(t)


def print_total_timer():
    """
    Print the content of the global TotalTimer, if it's not empty. This function will automatically get
    called when program exits.
    """
    if len(_TOTAL_TIMER_DATA) == 0:
        return
    for k, v in _TOTAL_TIMER_DATA.items():
        logger.info("Total Time: {} -> {:.2f} sec, {} times, {:.3g} sec/time".format(
            k, v.sum, v.count, v.average))


atexit.register(print_total_timer)


class IterSpeedCounter(object):
    """ Test how often some code gets reached.

    Example:
        Print the speed of the iteration every 100 times.

        .. code-block:: python

            speed = IterSpeedCounter(100)
            for k in range(1000):
                # do something
                speed()
    """

    def __init__(self, print_every, name=None):
        """
        Args:
            print_every(int): interval to print.
            name(str): name to used when print.
        """
        self.cnt = 0
        self.print_every = int(print_every)
        self.name = name if name else 'IterSpeed'

    def reset(self):
        self.start = timer()

    def __call__(self):
        if self.cnt == 0:
            self.reset()
        self.cnt += 1
        if self.cnt % self.print_every != 0:
            return
        t = timer() - self.start
        logger.info("{}: {:.2f} sec, {} times, {:.3g} sec/time".format(
            self.name, t, self.cnt, t / self.cnt))


class Timer():
    """
    A timer class which computes the time elapsed since the start/reset of the timer.
    """
    def __init__(self):
        self.reset()

    def reset(self):
        """
        Reset the timer.
        """
        self._start = timer()
        self._paused = False
        self._total_paused = 0

    def pause(self):
        """
        Pause the timer.
        """
        assert self._paused is False
        self._paused = timer()

    def is_paused(self):
        return self._paused is not False

    def resume(self):
        """
        Resume the timer.
        """
        assert self._paused is not False
        self._total_paused += timer() - self._paused
        self._paused = False

    def seconds(self):
        """
        Returns:
            float: the total number of seconds since the start/reset of the timer, excluding the
                time in between when the timer is paused.
        """
        if self._paused:
            self.resume()
            self.pause()
        return timer() - self._start - self._total_paused



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .task import BaseDataset, BaseTask, TaskData, TaskRunner, TransformerModel
from .task_glue import GlueData, GlueRunner, GlueTask
from .task_ner import NerData, NerRunner, NerTask



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# SPO数据集持久化，与标注文件之间的互转。

import json
import os
import re
from dataclasses import dataclass, field
from typing import List

from loguru import logger
from tqdm import tqdm


# 实体标签
@dataclass
class EntityTag:
    # 标签类别
    category: str = ""
    # 标签起始位置
    start: int = -1
    # 标签文本
    mention: str = ""

    def clear(self):
        pass

    def to_dict(self):
        return {
            'category': self.category,
            'start': self.start,
            'mention': self.mention
        }

    def from_dict(self, dict_data):
        self.clear()
        for k in self.__dict__.keys():
            if k in dict_data:
                v = dict_data[k]
                setattr(self, k, v)
        return self


# 关系标签
@dataclass
class RelTag:
    predicate: str = None
    sub: EntityTag = None
    obj: EntityTag = None

    def clear(self):
        self.predicate = None
        self.sub = None
        self.obj = None

    def to_dict(self):
        return {
            'predicate': self.predicate,
            'sub': self.sub.to_dict(),
            'obj': self.obj.to_dict()
        }

    def from_dict(self, dict_data):
        self.clear()
        self.predicate = dict_data.get('predicate', None)
        self.sub = dict_data.get('sub', None)
        if self.sub:
            self.sub = EntityTag().from_dict(self.sub)
        self.obj = dict_data.get('obj', None)
        if self.obj:
            self.obj = EntityTag().from_dict(self.obj)
        return self


# 打过标签的文本
@dataclass
class TaggedText:
    guid: str
    text: str
    spo_list: List[RelTag] = field(default_factory=list)

    def clear(self):
        self.spo_list = []

    def to_dict(self):
        return {
            'guid': self.guid,
            'text': self.text,
            'spo_list': [x.to_dict() for x in self.spo_list]
        }

    def from_dict(self, dict_data):
        self.clear()
        if 'guid' in dict_data:
            self.guid = dict_data['guid']
        elif 'text' in dict_data:
            self.text = dict_data['text']
        elif 'tags' in dict_data:
            for x in dict_data['spo_list']:
                self.spo_list.append(RelTag().from_dict(x))

        return self

    def add_tag(self, tag: RelTag):
        self.spo_list.append(tag)


# 实体关系标签数据集
class SpoDataset:
    def __init__(self, name=None, predicate_labels=None):
        self.name = name
        self.predicate_labels = predicate_labels
        self.tagged_text_list: List[TaggedText] = []

    def __len__(self):
        return len(self.tagged_text_list)

    def __getitem__(self, idx):
        return self.tagged_text_list[idx]

    def __iter__(self):
        for x in self.tagged_text_list:
            yield x

    def info(self):
        logger.info(f"Total: {len(self.tagged_text_list)}")

    def append(self, tagged_text: TaggedText):
        self.tagged_text_list.append(tagged_text)

    def extend(self, other_dataset):
        self.tagged_text_list.extend(other_dataset)

    def save(self, filename: str, format=None):
        """
        {'guid': '0000', 'text': "sample text", 'spo_list': [{'category': "实体类别1", start: 10, mention: "实体文本"}, ...]}
        """

        if format == "json":
            json_data = {x.guid: x.to_dict() for x in self.tagged_text_list}
            json.dump(json_data,
                      open(filename, 'w'),
                      ensure_ascii=False,
                      indent=2)
        else:
            with open(filename, 'w') as wt:
                for tagged_text in tqdm(self.tagged_text_list,
                                        desc="Save tagged text list."):
                    data_dict = tagged_text.to_dict()
                    wt.write(f"{json.dumps(data_dict, ensure_ascii=False)}\n")
        logger.info(f"Saved {filename}")

    def load_from_file(self, filename: str):
        logger.info(f"Loading {filename}")
        try:
            lines = [
                json.loads(x.strip()) for x in open(filename, 'r').readlines()
            ]
        except:
            logger.warning(
                f"{filename} is not standard style, try reviews style...")
            try:
                json_data = json.load(open(filename, 'r'))
                lines = [v for k, v in json_data.items()]
            except:
                logger.error("Unknown file style {filename}")
                raise Exception(f"Unknown file style {filename}")

        def data_generator(data_source=None):
            for json_data in lines:
                yield json_data['guid'], json_data['text'], None, json_data[
                    'tags']

        return self.load(data_generator)

    def load_from_brat_data(self, brat_data_dir):
        from .brat import brat_data_generator
        return self.load(brat_data_generator, brat_data_dir)

    def load(self, data_generator, data_source=None):
        for guid, text, _, json_tags in data_generator(data_source):
            tagged_text = TaggedText(guid, text)
            for json_tag in json_tags:
                #  logger.info(f"json_tag: {json_tag}")
                entity_tag = RelTag().from_dict(json_tag)
                tagged_text.add_tag(entity_tag)
            self.tagged_text_list.append(tagged_text)
        return self

    def export_to_brat(self, brat_data_dir, max_pages=10):
        from .brat import export_brat_files
        export_brat_files(self.tagged_text_list,
                          self.ner_labels,
                          self.ner_connections,
                          brat_data_dir,
                          max_pages=max_pages)

    def import_from_brat(self, brat_data_dir):
        from .brat import import_brat_files
        import_brat_files(self.tagged_text_list, brat_data_dir)

    def export_to_poplar(self, poplar_file, max_pages=100, start_page=0):
        from .poplar import save_poplar_file
        save_poplar_file(poplar_file,
                         self.ner_labels,
                         self.ner_connections,
                         start_page=start_page,
                         max_pages=max_pages)

    def import_from_poplar(self, poplar_file):
        from .poplar import poplar_data_generator
        return self.load(poplar_data_generator(poplar_file))

    def diff(self, another_ner_dataset):
        assert len(self) == len(another_ner_dataset)

        identical_tags_list = []
        a_only_tags_list = []
        b_only_tags_list = []
        for tagged_text_a, tagged_text_b in zip(self, another_ner_dataset):
            text_a, text_b = tagged_text_a.text, tagged_text_b.text
            # 被标注的文本必须相同，否则标注位置无意义
            assert text_a == text_b
            tags_a, tags_b = tagged_text_a.tags, tagged_text_b.tags
            tags_a = sorted(tags_a, key=lambda x: x.start)
            tags_b = sorted(tags_b, key=lambda x: x.start)

            identical_tags = []
            a_only_tags = []
            b_only_tags = []
            for tag_a in tags_a:
                found_a = False
                for tag_b in tags_b:
                    if tag_a == tag_b:
                        identical_tags.append(tag_a)
                        found_a = True
                        break
                if not found_a:
                    a_only_tags.append(tag_a)
            for tag_b in tags_b:
                found_b = False
                for tag_a in tags_a:
                    if tag_a == tag_b:
                        found_b = True
                        break
                if not found_b:
                    b_only_tags.append(tag_b)

            identical_tags_list.append(identical_tags)
            a_only_tags_list.append(a_only_tags)
            b_only_tags_list.append(b_only_tags)

        return identical_tags_list, a_only_tags_list, b_only_tags_list


def merge_spo_datasets(spo_dataset_list, min_dups=2):
    if len(spo_dataset_list) == 0:
        return None
    from copy import deepcopy
    merged_dataset = deepcopy(spo_dataset_list[0])
    if len(spo_dataset_list) > 1:
        for i, X in enumerate(zip(*spo_dataset_list)):
            spo_list_list = [x.spo_list for x in X]
            from ..utils import merge_entities
            new_spo_list = merge_entities(
                spo_list_list,
                key=lambda x: x.start * 1000 + len(x.mention),
                min_dups=min_dups)
            merged_dataset[i].spo_list = new_spo_list

    return merged_dataset


def spo_data_generator(train_file, dataset_name=None, predicate_labels=None):
    spo_dataset = SpoDataset(dataset_name, predicate_labels)
    spo_dataset.load_from_file(train_file)
    for tagged_text in tqdm(spo_dataset):
        guid = tagged_text.guid
        text = tagged_text.text
        spo_list = [x.to_dict() for x in tagged_text.spo_list]
        yield guid, text, None, spo_list


def load_spo_dataset(filename):
    spo_dataset = SpoDataset()
    spo_dataset.load_from_file(filename)
    return spo_dataset


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import random
import numpy as np


class Augmenter(object):
    """Data augmentation for the extractor.
    Support both token and span level augmentation operators.
    Attributes:
        index (dict): a dictionary containing both the token and span level index
        all_spans (dict): a dictionary from span type to a list of all the spans in the index
        span_freqs (dict): a dictionary from span type to the document frequency of each span in all_spans
    """
    def __init__(self, index_fn, valid_spans=None):
        self.index = json.load(open(index_fn))
        self.all_spans = {}
        self.span_freqs = {}

        for span_type in self.index['span']:
            self.all_spans[span_type] = list(
                self.index['span'][span_type].keys())
            span_freqs_tmp = [self.index['span'][span_type][sp]['document_freq'] \
                               for sp in self.all_spans[span_type]]
            self.span_freqs[span_type] = np.array(span_freqs_tmp) / np.sum(
                span_freqs_tmp)

    def augment(self, tokens, labels, op='token_del_tfidf'):
        """ Performs data augmentation on a tagging example.
        We support deletion (del), insertion (ins), replacement (repl),
        and swapping(swap) at the token level. At the span level, we support
        replacement with the options of replacing with random, frequent (freq),
        or similar spans (sim).
        The supported ops:
          ['token_del_tfidf',
           'token_del',
           'token_repl_tfidf',
           'token_repl',
           'token_swap',
           'token_ins',
           'span_sim',
           'span_freq',
           'span']
        Args:
            tokens (list of strings): the input tokens
            labels (list of strings): the labels of the tokens
            op (str, optional): a string encoding of the operator to be applied
        Returns:
            list of strings: the augmented tokens
            list of strings: the augmented labels
        """
        flags = op.split('_')
        if 'span' in flags:
            start, end = self.sample_span_position(tokens, labels)
            if start < 0:
                return tokens, labels
            span = ' '.join(tokens[start:end + 1])  # .lower()
            label = labels[start]
            if label.startswith('B-'):
                labelI = 'I' + labels[start][1:]
            else:
                labelI = label

            if 'AS' in label:
                span_type = 'aspect'
            else:
                span_type = 'opinion'

            if 'sim' in op:
                candidates = self.index['span'][span_type][span][
                    'similar_spans']
                new_span = random.choice(candidates)[0]
            elif 'freq' in op:
                candidates = self.all_spans[span_type]
                new_span = np.random.choice(candidates,
                                            1,
                                            p=self.span_freqs[span_type])[0]
            else:
                candidates = self.all_spans[span_type]
                new_span = random.choice(candidates)

            new_span_len = len(new_span.split(' '))
            new_tokens = tokens[:start] + \
                    new_span.split(' ') + tokens[end+1:]
            new_labels = labels[:start] + [label] + \
                [labelI] * (new_span_len - 1) + labels[end+1:]
            return new_tokens, new_labels
        else:
            tfidf = 'tfidf' in op
            pos1 = self.sample_position(tokens, labels, tfidf)
            if pos1 < 0:
                return tokens, labels

            if 'del' in op:
                # insert padding to keep the length consistent
                if tokens[pos1] in self.index['token']:
                    length = self.index['token'][tokens[pos1]]['bert_length']
                else:
                    length = 1
                new_tokens = tokens[:pos1] + ['[PAD]'
                                              ] * (length) + tokens[pos1 + 1:]
                new_labels = labels[:pos1] + ['<PAD>'
                                              ] * (length) + labels[pos1 + 1:]
            elif 'ins' in op:
                ins_token = self.sample_token(tokens[pos1], same_length=False)
                new_tokens = tokens[:pos1] + [ins_token] + tokens[pos1:]
                new_labels = labels[:pos1] + ['O'] + labels[pos1:]
            elif 'repl' in op:
                ins_token = self.sample_token(tokens[pos1], same_length=False)
                if tokens[pos1] in self.index['token'] and \
                      ins_token in self.index['token']:
                    len1 = self.index['token'][tokens[pos1]]['bert_length']
                    len2 = self.index['token'][ins_token]['bert_length']
                    if len1 < len2:
                        # truncate the new sequence
                        bert_tokens = self.index['token'][ins_token][
                            'bert_token'][:len1]
                        bert_tokens = [
                            token.replace('##', '') for token in bert_tokens
                        ]
                        ins_token = ''.join(bert_tokens)
                        new_tokens = tokens[:pos1] + [ins_token
                                                      ] + tokens[pos1 + 1:]
                        new_labels = labels[:pos1] + ['O'] + labels[pos1 + 1:]
                    else:
                        # pad the new sequence
                        more = len1 - len2
                        new_tokens = tokens[:pos1] + [
                            ins_token
                        ] + ['[PAD]'] * more + tokens[pos1 + 1:]
                        new_labels = labels[:pos1] + [
                            'O'
                        ] + ['<PAD>'] * more + labels[pos1 + 1:]
                else:
                    # backup
                    new_tokens = tokens[:pos1] + [ins_token
                                                  ] + tokens[pos1 + 1:]
                    new_labels = labels[:pos1] + ['O'] + labels[pos1 + 1:]
            elif 'swap' in op:
                pos2 = self.sample_position(tokens, labels, tfidf)
                new_tokens = list(tokens)
                new_labels = list(labels)
                new_tokens[pos1], new_tokens[pos2] = tokens[pos2], tokens[pos1]
            else:
                new_tokens, new_labels = tokens, labels

            return new_tokens, new_labels

    def augment_sent(self, text, op='token_del_tfidf'):
        """ Performs data augmentation on a classification example.
        Similar to augment(tokens, labels) but works for sentences
        or sentence-pairs.
        Args:
            text (str): the input sentence
            op (str, optional): a string encoding of the operator to be applied
        Returns:
            str: the augmented sentence
        """
        # handling sentence pairs
        sents = text.split(' [SEP] ')
        text = sents[0]
        target_spans = sents[1:]

        # tokenize the sentence
        current = ''
        tokens = []
        labels = []
        for ch in text:
            if ch.isalnum():
                current += ch
            else:
                if current != '':
                    tokens.append(current)
                if ch not in ' \t\r\n':
                    tokens.append(ch)
                current = ''
        if current != '':
            tokens.append(current)

        labels = ['O'] * len(tokens)
        for idx, span in enumerate(target_spans):
            span_tokens = span.split(' ')
            for tid in range(len(tokens)):
                if tid + len(span_tokens) <= len(tokens) and \
                   tokens[tid:tid+len(span_tokens)] == span_tokens:
                    for i in range(tid, tid + len(span_tokens)):
                        labels[i] = 'SP%d' % idx

        # print(tokens)
        # print(labels)
        # only augment the original sentence
        tokens, labels = self.augment(tokens, labels, op=op)

        # check consistency
        tid = 0
        while tid < len(tokens):
            if labels[tid][:2] == 'SP':
                new_span = tokens[tid]
                idx = int(labels[tid][2:])
                while tid + 1 < len(tokens) and \
                      labels[tid + 1] == labels[tid]:
                    tid += 1
                    new_span += ' ' + tokens[tid]
                if target_spans[idx] != new_span:
                    target_spans[idx] = new_span
            tid += 1

        # error handling
        results = ' '.join(tokens)
        for span in target_spans:
            results += ' [SEP] ' + span
        return results

    def sample_position(self, tokens, labels, tfidf=False):
        """ Randomly sample a token's position from a training example
        When tfidf is turned on, the weight of each token is proportional
        to MaxTfIdf - Tfidf of each token. When it is off, the sampling is uniform.
        Only tokens with 'O' labels and at least 1 position away from a non 'O'
        labels will be sampled.
        Args:
            tokens (list of strings): the input tokens
            labels (list of strings): the labels of the tokens
            tfidf (bool, optional): whether the sampled position is by tfidf
        Returns:
            int: the sampled position (-1 if no such position)
        """
        index = self.index['token']
        candidates = []
        for idx, token in enumerate(tokens):
            if labels[idx] == 'O' and \
                   token in index and \
                   index[token]['similar_words'] != None and \
                   len(index[token]['similar_words']) > 0:
                candidates.append(idx)
            # if token.lower() in index and \
            #     labels[idx] == 'O' and \
            #     (idx + 1 >= len(tokens) or labels[idx + 1] == 'O') and \
            #     (idx - 1 < 0 or labels[idx - 1] == 'O'):
            #     candidates.append(idx)

        if len(candidates) <= 0:
            return -1
        if tfidf:
            weight = {}
            max_weight = 0.0
            for idx, token in enumerate(tokens):
                # token = token.lower()
                if token not in index:
                    continue
                if token not in weight:
                    weight[token] = 0.0
                weight[token] += index[token]['idf']
                max_weight = max(max_weight, weight[token])

            weights = []
            for idx in candidates:
                weights.append(max_weight - weight[tokens[idx]] + 1e-6)
                # weights.append(max_weight - weight[tokens[idx].lower()] + 1e-6)
            weights = np.array(weights) / sum(weights)

            return np.random.choice(candidates, 1, p=weights)[0]
        else:
            return random.choice(candidates)

    def sample_token(self, token, same_length=True, max_candidates=10):
        """ Randomly sample a token's similar token stored in the index
        Args:
            token (str): the input token
            same_length (bool, optional): whether the return token should have the same
                length in BERT
            max_candidates (int, optional): the maximal number of candidates
                to be sampled
        Returns:
            str: the sampled token (unchanged if the input is not in index)
        """
        # token = token.lower()
        index = self.index['token']
        if token in index and index[token]['similar_words'] != None:
            if same_length:
                bert_length = index[token]['bert_length']
                candidates = []
                for ts, bl in zip(index[token]['similar_words'],
                                  index[token]['similar_words_length']):
                    t, _ = ts
                    if bl == bert_length:
                        candidates.append(t)
                        if len(candidates) >= max_candidates:
                            break
            else:
                candidates = [t for t, _ in \
                              index[token]['similar_words'][:max_candidates]]
            if len(candidates) <= 0:
                return token
            else:
                return random.choice(candidates)
        else:
            return token

    def sample_span_position(self, tokens, labels):
        """ Uniformly sample a span from a training example and return its positions.
        The output is a pair (start_op, end_op) of the span.
        Args:
            tokens (list of strings): the input tokens
            labels (list of strings): the labels of the tokens
        Returns:
            int: the start position (-1 if no available span)
            int: the ending position (-1 if no available span)
        """
        index = self.index['span']
        candidates = []
        idx = 0
        while idx < len(tokens):
            if labels[idx] != 'O':
                start = idx
                while idx + 1 < len(tokens) and \
                    labels[idx + 1][1:] == labels[idx][1:]:
                    idx += 1
                end = idx

                span = ' '.join(tokens[start:end + 1])  # .lower()
                if span in index:
                    candidates.append((start, end))
            idx += 1

        if len(candidates) > 0:
            return random.choice(candidates)
        else:
            return (-1, -1)


if __name__ == '__main__':
    from transformers import BertTokenizer
    tokenizer = BertTokenizer.from_pretrained('/opt/share/pretrained/pytorch/bert-base-uncased')

    ag = Augmenter('augment/laptop_index.json', [])
    tokens = 'this is a great drawback desktop keyboard'.split(' ')
    labels = 'O O O O O B-AS I-AS'.split(' ')
    for op in [
            'token_del_tfidf', 'token_del', 'token_repl_tfidf', 'token_repl',
            'token_swap', 'token_ins', 'span_sim', 'span_freq', 'span'
    ]:
        print(op)
        result = ag.augment(tokens, labels, op=op)
        result = ' '.join(result[0])
        print(result, len(tokenizer.encode(result)))
        original = ' '.join(tokens)
        print(original, len(tokenizer.encode(original)))

    tokens = 'I liked the macbook desktop keyboard . It is very good .'.split(
        ' ')
    labels = 'O O O O B-AS I-AS O O O O O O'.split(' ')
    for op in ['span']:
        print(op)
        print(ag.augment(tokens, labels, op=op))

    ag = Augmenter('augment/rest_index_asc.json', [])
    text = 'I liked the beef [SEP] beef'
    for op in [
            'token_del_tfidf', 'token_del', 'token_repl_tfidf', 'token_repl',
            'token_swap', 'token_ins', 'span_sim', 'span_freq', 'span'
    ]:
        print(op)
        print(ag.augment_sent(text, op=op))

    ag = Augmenter('augment/laptop_index_asc.json', [])
    text = 'I liked the desktop keyboard [SEP] desktop keyboard'
    for op in [
            'token_del_tfidf', 'token_del', 'token_repl_tfidf', 'token_repl',
            'token_swap', 'token_ins', 'span_sim', 'span_freq', 'span'
    ]:
        print(op)
        print(ag.augment_sent(text, op=op))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import json
import os
import random
from dataclasses import dataclass, field
from typing import List

from loguru import logger
from tqdm import tqdm

from theta.utils import seg_generator

from ..utils import DataClassBase

# ----------------------
# 实体标注规范
#  {
#      'guid':
#      "0",
#      'text':
#      'xxxxxxx',
#      'entities': [{
#          'category': category,
#          'start': start,
#          'end': end,
#          'mention': mention
#      }]
#  }


@dataclass(unsafe_hash=True)
class Entity(DataClassBase):
    category: str = field(default_factory=str)
    start: int = field(default_factory=int)
    end: int = field(default_factory=int)
    mention: str = field(default_factory=str)


@dataclass
class LabeledText(DataClassBase):
    guid: str = field(default_factory=str)
    text: str = field(default_factory=str)
    entities: List[Entity] = field(default_factory=list)

    def add_entity(self, category, start, end):
        self.entities.append(
            Entity(category=category,
                   start=start,
                   end=end,
                   mention=self.text[start:end + 1]))

    def offset(self, text_offset):
        for entity in self.entities:
            entity.start += text_offset
            entity.end += text_offset

    def get_annotated_text(self, seg_len, seg_backoff):
        #  debug = False
        #  if self.guid == '449':
        #      debug = True

        max_text_length = seg_len - seg_backoff

        text = self.text[:max_text_length]
        #  if debug:
        #      logger.debug(f"text: <{text}>")
        annotated_text = ""
        p0 = 0
        for entity in self.entities:
            x0 = entity.start
            x1 = entity.end
            if x1 >= max_text_length:
                continue
            annotated_text += text[p0:x0]
            #  if debug:
            #      logger.debug(
            #          f"annotated_text(p0:{p0},x0:{x0},x1:{x1}): {annotated_text}"
            #      )
            annotated_text += f"【{entity.mention}】"
            #  if debug:
            #      logger.debug(
            #          f"annotated_text(p0:{p0},x0:{x0},x1:{x1}): {annotated_text}"
            #      )
            p0 = x1 + 1
        annotated_text += text[p0:]
        #  if debug:
        #      logger.debug(f"annotated_text(p0:{p0}): {annotated_text}")

        return annotated_text

    #  def to_json(self):
    #      json_entities = [x.to_dict() for x in self.entities]
    #      json_data = {
    #          'guid': self.guid,
    #          'text': self.text,
    #          'entities': json_entities
    #      }
    #      return json_data

    #  def from_json(self, json_data):
    #      self.guid = json_data['guid']
    #      self.text = json_data['text']
    #      self.entities = []
    #      for x in json_data['entities']:
    #          entity = Entity()
    #          entity.from_dict(x)
    #          self.entities.append(entity)


def get_ner_preds_reviews(preds, examples, seg_len, seg_backoff):
    reviews = {}
    category_mentions = {}
    for json_d, input_example in tqdm(zip(preds, examples)):
        guid = input_example.guid
        text_offset = input_example.text_offset
        text = input_example.text_a
        entities = json_d['entities']

        #  logger.info(
        #      f"guid: {guid}, offset: {text_offset}, entities: {entities}")

        labeled_text = LabeledText(guid, text)
        for c, x0, x1 in entities:
            if x0 >= len(text):
                logger.info(f"x0({x0}) >= len(text)({len(text)})")
                logger.warning(
                    f"pos overflow [{guid}]:({c},{x0},{x1}) text: ({len(text)}) {text}"
                )
            if x0 < 0 or x1 < 0 or x0 > x1:
                logger.warning(f"Invalid x0({x0}) and x1({x1})")

            labeled_text.add_entity(c, x0, x1)
            if c not in category_mentions:
                category_mentions[c] = set()
            category_mentions[c].add(labeled_text.entities[-1].mention)

        annotated_text = labeled_text.get_annotated_text(seg_len, seg_backoff)

        labeled_text.offset(text_offset)
        json_data = labeled_text.to_dict()
        json_entities = json_data['entities']

        if guid not in reviews:
            reviews[guid] = {
                'guid': guid,
                'text': "",
                'annotated_text': "",
                'tags': []
            }
        reviews[guid]['text'] += text[:seg_len - seg_backoff]
        reviews[guid]['annotated_text'] += annotated_text
        reviews_tags = reviews[guid]['tags']
        for x in json_entities:
            if x not in reviews_tags:
                c = x['category']
                s = int(x['start'])
                m = x['mention']
                #  e = s + len(m) - 1
                #  m = text[s:e + 1]

                #  if not m:
                #      logger.warning(
                #          f"Mention is None. guid: {guid}, text: {text}, c: {c}, s: {s}, m: {m}"
                #      )
                reviews_tags.append({'category': c, 'start': s, 'mention': m})

        reviews_tags = sorted(reviews_tags, key=lambda x: x['start'])
        from ..utils import remove_duplicate_entities
        reviews_tags = remove_duplicate_entities(reviews_tags)
        reviews[guid]['tags'] = reviews_tags
        #  logger.info(f"guid: {guid}, tags: {reviews[guid]['tags']}")

    return reviews, category_mentions


def save_ner_preds(args, preds, test_examples):
    reviews, category_mentions = get_ner_preds_reviews(preds, test_examples,
                                                       args.seg_len,
                                                       args.seg_backoff)

    #  reviews_file = f"{args.latest_dir}/{args.dataset_name}_reviews_{args.local_id}.json"
    reviews_file = args.reviews_file

    json.dump(reviews, open(reviews_file, 'w'), ensure_ascii=False, indent=2)
    logger.info(f"Reviews file: {reviews_file}")

    category_mentions_file = f"{args.latest_dir}/{args.dataset_name}_category_mentions_{args.local_id}.txt"
    num_categories = len(category_mentions)
    num_mentions = 0
    with open(category_mentions_file, 'w') as wt:
        for c, mentions in category_mentions.items():
            for m in mentions:
                wt.write(f"{c}\t{m}\n")
                num_mentions += 1
    logger.info(
        f"Total {num_categories} categories and {num_mentions} mentions saved to {category_mentions_file}"
    )

    #  ----- Tracking -----
    #  if args.do_experiment:
    #      mlflow.log_param(f"{args.dataset_name}_reviews_file", reviews_file)
    #      mlflow.log_artifact(reviews_file, args.artifact_path)
    #      mlflow.log_param(f"{args.dataset_name}_category_mentions_file",
    #                       category_mentions_file)
    #      mlflow.log_artifact(category_mentions_file, args.artifact_path)

    return reviews_file, category_mentions_file


def augment_entities(all_text_entities, labels_map, num_augments):
    aug_tokens = []
    for i, (guid, text, entities) in enumerate(
            tqdm(all_text_entities, desc=f"Augement {num_augments}X")):

        #  print(f"-------------------{json_file}--------------------")
        #  print(text)
        #  print(entities)
        #  for entity in entities:
        #      s = entity['start_pos']
        #      e = entity['end_pos']
        #      print(f"{entity['label_type']}: {text[s:e]}")
        #  print("----------------------------------------")
        if entities:
            for ai in range(num_augments):
                e_idx = random.randint(0, len(entities) - 1)
                entity = entities[e_idx]

                label_type = entity['category']
                s = entity['start']
                e = entity['end']

                labels = labels_map[label_type]
                idx = random.randint(0, len(labels) - 1)
                new_entity_text = labels[idx]

                text = text[:s] + new_entity_text + text[e + 1:]

                assert len(new_entity_text) >= 0
                delta = len(new_entity_text) - (e + 1 - s)

                entity['end'] = entity['start'] + len(new_entity_text) - 1
                entity['mention'] = new_entity_text

                #  assert text[entity['start']:entity['end'] +
                #              1] == new_entity_text
                text_mention = text[entity['start']:entity['end'] + 1]
                if text_mention != new_entity_text:
                    logger.warning(
                        f"Augment missing: |||{new_entity_text}||| vs |||{text_mention}|||"
                    )

                for n, e in enumerate(entities):
                    if n > e_idx:
                        e['start'] += delta
                        e['end'] += delta

                aug_tokens.append(
                    (f"{guid}-a{ai}", text, copy.deepcopy(entities)))

    #  for guid, text, entities in aug_tokens:
    #      text_a = text
    #      for entity in entities:
    #          logger.debug(f"{guid}: text_a: {text_a}")
    #          logger.debug(
    #              f"text_a[entity['start']:entity['end']]: {text_a[entity['start']:entity['end']]}"
    #          )
    #          logger.debug(
    #              f"mention {entity['mention']} in {text_a.find(entity['mention'])}"
    #          )
    #          logger.debug(f"entity: {entity}")
    #          assert text_a[entity['start']:entity['end']] == entity[
    #              'mention']

    return aug_tokens


def data_seg_generator(lines,
                       ner_labels,
                       seg_len=0,
                       seg_backoff=0,
                       num_augments=0,
                       allow_overlap=False):
    #  assert seg_backoff >= 0 and seg_backoff <= int(seg_len * 3 / 4)
    assert seg_len >= 0 and seg_backoff >= 0 and seg_backoff <= seg_len
    all_text_entities = []
    labels_map = {}

    logger.warning(f"{len(lines) in lines}")
    num_overlap = 0
    for i, s in enumerate(tqdm(lines)):
        guid = str(i)
        text = s['text']
        entities = s['entities']

        new_entities = []
        used_span = []
        #  logger.debug(f"entities: {entities}")
        entities = sorted(entities, key=lambda e: e.start)
        for entity in entities:
            if entity.category not in ner_labels:
                continue
            s = entity.start
            e = entity.end
            entity.mention = text[s:e + 1]

            overlap = False
            for us in used_span:
                if s > us[0] and s <= us[1] and e > us[1]:
                    overlap = True
                    logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                    break
                if e >= us[0] and e < us[1] and s < us[0]:
                    overlap = True
                    logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                    break

                #  if s >= us[0] and s <= us[1]:  # and e > us[1]:
                #      overlap = True
                #      logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                #      break
                #  if e >= us[0] and e <= us[1]:  # and s < us[0]:
                #      overlap = True
                #      logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                #      break
                #  if us[0] >= s and us[0] <= e:  # and e > us[1]:
                #      overlap = True
                #      logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                #      break
                #  if us[1] >= s and us[1] <= e:  # and s < us[0]:
                #      overlap = True
                #      logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                #      break

                #  if s >= us[0] and s <= us[1]:
                #      overlap = True
                #      logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                #      break
                #  if e >= us[0] and e <= us[1]:
                #      overlap = True
                #      logger.warning(f"Overlap: ({s},{e}) vs ({us[0],us[1]})")
                #      break
            if overlap:
                num_overlap += 1
                #  if not allow_overlap:
                logger.warning(
                    f"Overlap! {i} mention: {entity.mention}({s}:{e}), used_span: {used_span}"
                )
                continue
            used_span.append((s, e))

            new_entities.append(entity)
        #  logger.warning(f"{len(new_entities)} new_entities")
        entities = new_entities

        seg_offset = 0
        #  if seg_len <= 0:
        #      seg_len = max_seq_length

        for (seg_text, ), text_offset in seg_generator((text, ), seg_len,
                                                       seg_backoff):
            text_a = seg_text

            assert text_offset == seg_offset

            seg_start = seg_offset
            seg_end = seg_offset + min(seg_len, len(seg_text))
            labels = [(x.category, x.start - seg_offset, x.end - seg_offset)
                      for x in entities
                      if x.start >= seg_offset and x.end < seg_end]

            # 没有标注存在的文本片断不用于训练
            if labels:
                yield guid, text_a, None, labels, seg_offset

            #      if num_augments > 0:
            #          seg_entities = [
            #              {
            #                  'start': x.start - seg_offset,
            #                  'end': x.end - seg_offset,
            #                  'category': x.category,
            #                  'mention': x.mention
            #              } for x in entities
            #              if x.start >= seg_offset and x.end < seg_end
            #          ]
            #          all_text_entities.append((guid, text_a, seg_entities))
            #
            #          for entity in seg_entities:
            #              label_type = entity['category']
            #              s = entity['start']  # - seg_offset
            #              e = entity['end']  #- seg_offset
            #              #  print(s, e)
            #              assert e >= s
            #              #  logger.debug(
            #              #      f"seg_start: {seg_start}, seg_end: {seg_end}, seg_offset: {seg_offset}"
            #              #  )
            #              #  logger.debug(f"s: {s}, e: {e}")
            #              assert s >= 0 and e < len(seg_text)
            #              #  if s >= len(seg_text) or e >= len(seg_text):
            #              #      continue
            #
            #              entity_text = seg_text[s:e + 1]
            #              #  print(label_type, entity_text)
            #
            #              assert len(entity_text) > 0
            #              if label_type not in labels_map:
            #                  labels_map[label_type] = []
            #              labels_map[label_type].append(entity_text)

            seg_offset += seg_len - seg_backoff

    logger.warning(f"num_overlap: {num_overlap}")

    if num_augments > 0:
        aug_tokens = augment_entities(all_text_entities,
                                      labels_map,
                                      num_augments=num_augments)
        for guid, text, entities in aug_tokens:
            text_a = text
            for entity in entities:
                #  logger.debug(f"text_a: {text_a}")
                #  logger.debug(
                #      f"text_a[entity['start']:entity['end']]: {text_a[entity['start']:entity['end']]}"
                #  )
                #  logger.debug(
                #      f"mention {entity['mention']} in {text_a.find(entity['mention'])}"
                #  )
                #  logger.debug(f"entity: {entity}")

                #  assert text_a[entity['start']:entity['end'] + 1] == entity['mention']
                pass

            labels = [
                (entity['category'], entity['start'], entity['end'])
                for entity in entities if entity['end'] < (
                    min(len(text_a), seg_len) if seg_len > 0 else len(text_a))
            ]
            yield guid, text_a, None, labels, 0


class InputExample(object):
    """A single training/test example for token classification."""
    def __init__(self, guid, text_a, labels, text_offset=0):
        self.guid = guid
        self.text_a = text_a
        self.labels = labels
        self.text_offset = text_offset

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


def load_ner_examples(data_generator, examples_file, seg_len=0, seg_backoff=0):

    examples = []

    for guid, text_a, _, labels in data_generator(examples_file):
        assert text_a is not None
        for (seg_text_a, ), text_offset in seg_generator((text_a, ), seg_len,
                                                         seg_backoff):
            examples.append(
                InputExample(guid=guid,
                             text_a=seg_text_a,
                             labels=labels,
                             text_offset=text_offset))
    logger.info(f"Loaded {len(examples)} examples.")

    return examples


def load_ner_labeled_examples(lines,
                              ner_labels,
                              seg_len=0,
                              seg_backoff=0,
                              num_augments=0,
                              allow_overlap=False):

    examples = []
    for guid, text_a, text_b, labels, text_offset in data_seg_generator(
            lines,
            ner_labels,
            seg_len=seg_len,
            seg_backoff=seg_backoff,
            num_augments=num_augments,
            allow_overlap=allow_overlap):
        assert text_a is not None
        examples.append(
            InputExample(guid=guid,
                         text_a=text_a,
                         labels=labels,
                         text_offset=text_offset))

    logger.info(f"Loaded {len(examples)} examples.")
    return examples


def show_ner_datainfo(ner_labels, train_data_generator, train_file,
                      test_data_generator, test_file):
    from collections import defaultdict
    label_counts = defaultdict(int)

    label_examples = defaultdict(int)

    train_lengths = []
    for guid, text_a, _, labels in train_data_generator(train_file):
        entity_examples = defaultdict(int)
        train_lengths.append(len(text_a))
        for entity in labels:
            c = entity['category']
            entity_examples[c] = 1
            label_counts[c] += 1
            if c not in ner_labels:
                ner_labels.append(c)
        for c, n in entity_examples.items():
            label_examples[c] += n

    test_lengths = [
        len(text_a)
        for guid, text_a, _, labels in test_data_generator(test_file)
    ]

    logger.info(f"****** ner_labels ******")
    logger.info(f"{len(ner_labels)} labels: {ner_labels}")
    logger.info(f"label_counts: {label_counts}")
    logger.info(f"label_examples: {label_examples}")
    logger.info(f"")
    import numpy as np
    logger.warning(
        f"num train examples: {len(train_lengths)}, num test examples: {len(test_lengths)}"
    )
    logger.info(f"****** train lengths ******")
    logger.info(f"mean: {np.mean(train_lengths):.2f}")
    logger.info(f"std: {np.std(train_lengths):.2f}")
    logger.info(f"min: {np.min(train_lengths)}")
    logger.info(f"max: {np.max(train_lengths)}")
    logger.info(f"")

    logger.info(f"****** test lengths ******")
    logger.info(f"mean: {np.mean(test_lengths):.2f}")
    logger.info(f"std: {np.std(test_lengths):.2f}")
    logger.info(f"min: {np.min(test_lengths)}")
    logger.info(f"max: {np.max(test_lengths)}")
    logger.info(f"")


def to_poplar(args, poplar_data_file, pages, ner_labels, ner_connections,
              start_page, max_pages):
    poplar_json = {
        "content": "",
        "labelCategories": [],
        "labels": [],
        "connectionCategories": [],
        "connections": []
    }

    poplar_colorset = [
        '#007bff', '#17a2b8', '#28a745', '#fd7e14', '#e83e8c', '#dc3545',
        '#20c997', '#ffc107', '#007bff'
    ]
    label2id = {x: i for i, x in enumerate(ner_labels)}
    label_categories = poplar_json['labelCategories']
    for _id, x in enumerate(ner_labels):
        label_categories.append({
            "id":
            _id,
            "text":
            x,
            "color":
            poplar_colorset[label2id[x] % len(poplar_colorset)],
            "borderColor":
            "#cccccc"
        })

    connection_categories = poplar_json['connectionCategories']
    for _id, _text in enumerate(ner_connections):
        connection_categories.append({'id': _id, 'text': _text})

    poplar_labels = poplar_json['labels']
    poplar_content = ""
    eid = 0
    num_pages = 0
    page_offset = 0

    for guid, text in pages:

        if num_pages < start_page:
            num_pages += 1
            continue

        page_head = f"\n-------------------- {guid} Begin --------------------\n\n"
        page_tail = f"\n-------------------- {guid} End --------------------\n\n"
        poplar_content += page_head + f"{text}" + page_tail

        num_pages += 1
        if num_pages - start_page >= max_pages:
            break

        page_offset = len(poplar_content)

    poplar_json["content"] = poplar_content
    poplar_json['labels'] = poplar_labels

    json.dump(poplar_json,
              open(poplar_data_file, 'w'),
              ensure_ascii=False,
              indent=2)
    logger.info(f"Saved {poplar_data_file}")


def to_sampling_poplar(args,
                       train_data_generator,
                       ner_labels,
                       ner_connections,
                       start_page=0,
                       max_pages=100):
    poplar_json = {
        "content": "",
        "labelCategories": [],
        "labels": [],
        "connectionCategories": [],
        "connections": []
    }

    poplar_colorset = [
        '#007bff', '#17a2b8', '#28a745', '#fd7e14', '#e83e8c', '#dc3545',
        '#20c997', '#ffc107', '#007bff'
    ]
    label2id = {x: i for i, x in enumerate(ner_labels)}
    label_categories = poplar_json['labelCategories']
    for _id, x in enumerate(ner_labels):
        label_categories.append({
            "id":
            _id,
            "text":
            x,
            "color":
            poplar_colorset[label2id[x] % len(poplar_colorset)],
            "borderColor":
            "#cccccc"
        })

    connection_categories = poplar_json['connectionCategories']
    for _id, _text in enumerate(ner_connections):
        connection_categories.append({'id': _id, 'text': _text})

    poplar_labels = poplar_json['labels']
    poplar_content = ""
    eid = 0
    num_pages = 0
    page_offset = 0
    for guid, text, _, entities in train_data_generator(args.train_file):
        if num_pages < start_page:
            num_pages += 1
            continue

        page_head = f"\n-------------------- {guid} Begin --------------------\n\n"
        page_tail = f"\n-------------------- {guid} End --------------------\n\n"
        poplar_content += page_head + f"{text}" + page_tail

        for entity in entities:
            poplar_labels.append({
                "id":
                eid,
                "categoryId":
                label2id[entity.category],
                "startIndex":
                page_offset + len(page_head) + entity.start,
                "endIndex":
                page_offset + len(page_head) + entity.end + 1,
            })
            eid += 1

        num_pages += 1
        if num_pages - start_page >= max_pages:
            break

        page_offset = len(poplar_content)

    poplar_json["content"] = poplar_content
    poplar_json['labels'] = poplar_labels

    if not os.path.exists("./poplar"):
        os.makedirs("./poplar")
    poplar_data_file = f"./poplar/poplar_sampling_data_{args.local_id}_{max_pages}_{start_page}.json"
    json.dump(poplar_json,
              open(poplar_data_file, 'w'),
              ensure_ascii=False,
              indent=2)
    logger.info(f"Saved {poplar_data_file}")


def to_train_poplar(args,
                    train_data_generator,
                    ner_labels,
                    ner_connections,
                    start_page=0,
                    max_pages=100):
    poplar_json = {
        "content": "",
        "labelCategories": [],
        "labels": [],
        "connectionCategories": [],
        "connections": []
    }

    poplar_colorset = [
        '#007bff', '#17a2b8', '#28a745', '#fd7e14', '#e83e8c', '#dc3545',
        '#20c997', '#ffc107', '#007bff'
    ]
    label2id = {x: i for i, x in enumerate(ner_labels)}
    label_categories = poplar_json['labelCategories']
    for _id, x in enumerate(ner_labels):
        label_categories.append({
            "id":
            _id,
            "text":
            x,
            "color":
            poplar_colorset[label2id[x] % len(poplar_colorset)],
            "borderColor":
            "#cccccc"
        })

    connection_categories = poplar_json['connectionCategories']
    for _id, _text in enumerate(ner_connections):
        connection_categories.append({'id': _id, 'text': _text})

    poplar_labels = poplar_json['labels']
    poplar_content = ""
    eid = 0
    num_pages = 0
    page_offset = 0
    for guid, text, _, entities in train_data_generator(args.train_file):
        if num_pages < start_page:
            num_pages += 1
            continue

        page_head = f"\n-------------------- {guid} Begin --------------------\n\n"
        page_tail = f"\n-------------------- {guid} End --------------------\n\n"
        poplar_content += page_head + f"{text}" + page_tail

        for entity in entities:
            poplar_labels.append({
                "id":
                eid,
                "categoryId":
                label2id[entity.category],
                "startIndex":
                page_offset + len(page_head) + entity.start,
                "endIndex":
                page_offset + len(page_head) + entity.end + 1,
            })
            eid += 1

        num_pages += 1
        if num_pages - start_page >= max_pages:
            break

        page_offset = len(poplar_content)

    poplar_json["content"] = poplar_content
    poplar_json['labels'] = poplar_labels

    if not os.path.exists("./poplar"):
        os.makedirs("./poplar")
    poplar_data_file = f"./poplar/poplar_train_data_{args.local_id}_{max_pages}_{start_page}.json"
    json.dump(poplar_json,
              open(poplar_data_file, 'w'),
              ensure_ascii=False,
              indent=2)
    logger.info(f"Saved {poplar_data_file}")


def to_reviews_poplar(args,
                      ner_labels,
                      ner_connections,
                      start_page=0,
                      max_pages=100):
    poplar_json = {
        "content": "",
        "labelCategories": [],
        "labels": [],
        "connectionCategories": [],
        "connections": []
    }

    poplar_colorset = [
        '#007bff', '#17a2b8', '#28a745', '#fd7e14', '#e83e8c', '#dc3545',
        '#20c997', '#ffc107', '#007bff'
    ]
    label2id = {x: i for i, x in enumerate(ner_labels)}
    label_categories = poplar_json['labelCategories']
    for _id, x in enumerate(ner_labels):
        label_categories.append({
            "id":
            _id,
            "text":
            x,
            "color":
            poplar_colorset[label2id[x] % len(poplar_colorset)],
            "borderColor":
            "#cccccc"
        })

    connection_categories = poplar_json['connectionCategories']
    for _id, _text in enumerate(ner_connections):
        connection_categories.append({'id': _id, 'text': _text})

    poplar_labels = poplar_json['labels']
    poplar_content = ""
    num_pages = 0
    page_offset = 0

    reviews_file = f"{args.latest_dir}/{args.dataset_name}_reviews_{args.local_id}.json"
    reviews = json.load(open(reviews_file, 'r'))

    eid = 0
    for guid, json_data in reviews.items():
        if num_pages < start_page:
            num_pages += 1
            continue
        text = json_data['text']

        page_head = f"\n-------------------- {guid} Begin --------------------\n\n"
        page_tail = f"\n-------------------- {guid} End --------------------\n\n"
        poplar_content += page_head + f"{text}" + page_tail

        for json_entity in json_data['entities']:
            category = json_entity['category']
            mention = json_entity['mention']
            start_char = json_entity['start']
            end_char = json_entity['end']

            poplar_labels.append({
                "id":
                eid,
                "categoryId":
                label2id[category],
                "startIndex":
                page_offset + len(page_head) + start_char,
                "endIndex":
                page_offset + len(page_head) + end_char + 1,
            })
            eid += 1

        num_pages += 1
        if num_pages - start_page >= max_pages:
            break

        page_offset = len(poplar_content)

    poplar_json["content"] = poplar_content
    poplar_json['labels'] = poplar_labels

    if not os.path.exists("./poplar"):
        os.makedirs("./poplar")
    poplar_data_file = f"./poplar/poplar_reviews_data_{args.local_id}_{max_pages}_{start_page}.json"
    json.dump(poplar_json,
              open(poplar_data_file, 'w'),
              ensure_ascii=False,
              indent=2)
    logger.info(f"Saved {poplar_data_file}")


def load_train_val_examples(args,
                            train_data_generator,
                            ner_labels,
                            shuffle=True,
                            train_rate=0.9,
                            num_augments=0,
                            aug_train_only=False):
    logger.warning(
        f"shuffle: {shuffle}, train_rate: {train_rate:.2f}, num_augments: {num_augments}"
    )
    #  from theta.modeling import LabeledText, load_ner_labeled_examples
    lines = []
    #  for guid, text, _, entities in train_data_generator(args.train_file):
    #      sl = LabeledText(guid, text, entities)
    #      lines.append({'guid': guid, 'text': text, 'entities': entities})

    for guid, text, _, tags in train_data_generator(args.train_file):
        sl = LabeledText(guid, text)
        for tag in tags:
            c = tag['category']
            s = tag['start']
            m = tag['mention']
            if len(m) == 0:
                continue
            sl.add_entity(c, s, s + len(m) - 1)
        lines.append({'guid': guid, 'text': text, 'entities': sl.entities})
        if args.max_train_examples > 0 and len(
                lines) >= args.max_train_examples:
            break

    allow_overlap = args.allow_overlap
    if num_augments > 0:
        allow_overlap = False

    logger.warning(f"len(lines): {len(lines)}")

    train_base_examples = load_ner_labeled_examples(
        lines,
        ner_labels,
        seg_len=args.seg_len,
        seg_backoff=args.seg_backoff,
        num_augments=num_augments,
        allow_overlap=allow_overlap)

    if args.train_sample_rate < 1.0:
        num_samples = int(len(train_base_examples) * args.train_sample_rate)
        train_base_examples = train_base_examples[:num_samples]

    logger.warning(f"len(train_base_examples): {len(train_base_examples)}")

    from ..utils import split_train_eval_examples
    train_examples, val_examples = split_train_eval_examples(
        train_base_examples,
        train_rate=train_rate,
        fold=args.fold,
        shuffle=shuffle)

    logger.info(f"Loaded {len(train_examples)} train examples, "
                f"{len(val_examples)} val examples.")
    return train_examples, val_examples


#  def load_train_val_examples(args,
#                              train_data_generator,
#                              ner_labels,
#                              shuffle=True,
#                              train_rate=0.9,
#                              num_augments=0,
#                              aug_train_only=False):
#      logger.warning(
#          f"shuffle: {shuffle}, train_rate: {train_rate:.2f}, num_augments: {num_augments}"
#      )
#      #  from theta.modeling import LabeledText, load_ner_labeled_examples
#      lines = []
#      #  for guid, text, _, entities in train_data_generator(args.train_file):
#      #      sl = LabeledText(guid, text, entities)
#      #      lines.append({'guid': guid, 'text': text, 'entities': entities})
#
#      for guid, text, _, tags in train_data_generator(args.train_file):
#          sl = LabeledText(guid, text)
#          for tag in tags:
#              c = tag['category']
#              s = tag['start']
#              m = tag['mention']
#              if len(m) == 0:
#                  continue
#              sl.add_entity(c, s, s + len(m) - 1)
#          lines.append({'guid': guid, 'text': text, 'entities': sl.entities})
#          if args.max_train_examples > 0 and len(
#                  lines) >= args.max_train_examples:
#              break
#
#      allow_overlap = args.allow_overlap
#      if num_augments > 0:
#          allow_overlap = False
#
#      num_lines = len(lines)
#      num_samples = num_lines
#      if args.train_sample_rate < 1.0:
#          num_samples = int(len(train_base_examples) * args.train_sample_rate)
#          lines = lines[:num_samples]
#      logger.warning(
#          f"Train sample rate: {args.train_sample_rate:.2f}, use {num_samples}/{num_lines} examples."
#      )
#
#      from ..utils import split_train_eval_examples
#      if aug_train_only:
#          train_lines, val_lines = split_train_eval_examples(
#              lines, train_rate=train_rate, fold=args.fold, shuffle=shuffle)
#
#          train_examples = load_ner_labeled_examples(
#              train_lines,
#              ner_labels,
#              seg_len=args.seg_len,
#              seg_backoff=args.seg_backoff,
#              num_augments=num_augments,
#              allow_overlap=allow_overlap)
#
#          val_examples = load_ner_labeled_examples(val_lines,
#                                                   ner_labels,
#                                                   seg_len=args.seg_len,
#                                                   seg_backoff=args.seg_backoff,
#                                                   num_augments=0,
#                                                   allow_overlap=allow_overlap)
#      else:
#          train_base_examples = load_ner_labeled_examples(
#              lines,
#              ner_labels,
#              seg_len=args.seg_len,
#              seg_backoff=args.seg_backoff,
#              num_augments=num_augments,
#              allow_overlap=allow_overlap)
#
#          #  if args.train_sample_rate < 1.0:
#          #      num_samples = int(
#          #          len(train_base_examples) * args.train_sample_rate)
#          #      train_base_examples = train_base_examples[:num_samples]
#
#          logger.warning(f"len(train_base_examples): {len(train_base_examples)}")
#
#          train_examples, val_examples = split_train_eval_examples(
#              train_base_examples,
#              train_rate=train_rate,
#              fold=args.fold,
#              shuffle=shuffle)
#
#      logger.info(f"Loaded {len(train_examples)} train examples, "
#                  f"{len(val_examples)} val examples.")
#      return train_examples, val_examples


def load_test_examples(args, test_data_generator):
    #  from theta.modeling import load_ner_examples
    test_base_examples = load_ner_examples(test_data_generator,
                                           args.test_file,
                                           seg_len=args.seg_len,
                                           seg_backoff=args.seg_backoff)

    logger.info(f"Loaded {len(test_base_examples)} test examples.")
    return test_base_examples


def ner_evaluate(dev_file, reviews_file, eval_data_generator):
    dev_data = []
    for guid, text, _, tags in eval_data_generator(dev_file):
        pos = [(x['start'], x['start'] + len(x['mention'])) for x in tags]
        pos = sorted(pos)
        #  logger.info(f"{pos}")
        dev_data.append(pos)

    reviews_data = []
    from theta.modeling import ner_data_generator
    for guid, text, _, tags in ner_data_generator(reviews_file):
        pos = [(x['start'], x['start'] + len(x['mention'])) for x in tags]
        pos = sorted(pos)
        #  logger.info(f"{pos}")
        reviews_data.append(pos)

    total_right = 0
    total_preds = 0
    total_targets = 0
    P = 0.0
    R = 0.0
    F1 = 0.0
    for pred, target in tqdm(zip(reviews_data, dev_data)):
        num_preds = len(pred)
        num_targets = len(target)
        right = set(pred) & set(target)
        num_right = len(right)

        total_right += num_right
        total_preds += num_preds
        total_targets += num_targets

        if num_preds == 0:
            p = 0.0
            r = 0.0
        else:
            p = num_right / len(pred)
            r = num_right / len(target)
        if p == 0.0 and r == 0.0:
            f1 = 0.0
        else:
            f1 = 2 * p * r / (p + r)
        P += p
        R += r
        F1 += f1
    logger.info(
        f"total_right: {total_right}, total_preds: {total_preds}, total_targets: {total_targets}"
    )
    micro_f1 = F1 / len(dev_data)
    micro_p = P / len(dev_data)
    micro_r = R / len(dev_data)

    logger.warning(
        f"Micro: P: {micro_p:.6f}, R: {micro_r:.6f}, F1: {micro_f1:.6f}")

    macro_p = total_right / total_preds
    macro_r = total_right / total_targets
    macro_f1 = 2 * macro_p * macro_r / (macro_p + macro_r)
    logger.warning(
        f"Macro: P: {macro_p:.6f}, R: {macro_r:.6f}, F1: {macro_f1:.6f}")

    eval_results = {
        'macro_p': macro_p,
        'macro_r': macro_r,
        'macro_f1': macro_f1,
        'micro_p': micro_p,
        'micro_r': micro_r,
        'micro_f1': micro_f1,
    }
    return eval_results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
PointerSequenceTagger

指针序列标注器

"""

from copy import deepcopy
import json

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from loguru import logger
from tqdm import tqdm

from .base_tagger import BaseTagger
from .tagger_utils import batch_encode_texts, batch_encode_tags, common_to_tensors


class InputFeature(object):
    """A single set of features of data."""
    def __init__(self, input_ids, attention_mask, token_type_ids, input_len,
                 token_offsets, start_ids, end_ids, encoded_tags, text):
        self.input_ids = input_ids
        self.attention_mask = attention_mask
        self.token_type_ids = token_type_ids
        self.input_len = input_len
        self.token_offsets = token_offsets

        self.start_ids = start_ids
        self.end_ids = end_ids
        self.encoded_tags = encoded_tags
        self.text = text

    def __repr__(self):
        return json.dumps(self.to_dict(),
                          ensure_asci=False,
                          indent=2,
                          sort_keys=True)

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()

    def to_dict(self):
        return deepcopy(self.__dict__)


def encode_examples(examples, tokenizer, label2id, max_seq_length):
    """
    :param 
    """

    # ------------------------------ encode all texts ------------------------------
    all_texts = [e.text for e in examples]
    all_tokens, all_token2char, all_char2token, all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets = batch_encode_texts(
        all_texts, label2id, tokenizer, max_seq_length)

    # ------------------------------ encode all tags ------------------------------
    all_tags = [e.tags for e in examples]
    all_token_lens = [len(input_ids) for input_ids in all_input_ids]
    all_start_ids, all_end_ids, all_encoded_tags = batch_encode_tags(
        all_tags, all_token_lens, all_char2token, label2id, max_seq_length)

    # ------------------------------ check ------------------------------
    logger.debug(f"all_input_ids.shape: {np.array(all_input_ids).shape}")
    logger.debug(
        f"all_attention_mask.shape: {np.array(all_attention_mask).shape}")
    logger.debug(
        f"all_token_type_ids.shape: {np.array(all_token_type_ids).shape}")
    logger.debug(f"all_start_ids.shape: {np.array(all_start_ids).shape}")
    logger.debug(f"all_end_ids.shape: {np.array(all_end_ids).shape}")
    logger.debug(f"all_encoded_tags.shape: {np.array(all_encoded_tags).shape}")
    logger.debug(f"all_input_lens.shape: {np.array(all_input_lens).shape}")
    assert np.array(all_input_ids).shape[1] == max_seq_length
    assert np.array(all_attention_mask).shape[1] == max_seq_length
    assert np.array(all_token_type_ids).shape[1] == max_seq_length

    # ------------------------------ convert to tensor ------------------------------
    all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets = common_to_tensors(
        all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens,
        all_token_offsets)

    all_start_ids = torch.from_numpy(np.array(all_start_ids, dtype=np.int64))
    all_end_ids = torch.from_numpy(np.array(all_end_ids, dtype=np.int64))

    # ------------------------------ build features ------------------------------
    all_features = [
        InputFeature(input_ids=input_ids,
                     attention_mask=attention_mask,
                     token_type_ids=token_type_ids,
                     input_len=input_len,
                     token_offsets=token_offsets,
                     start_ids=start_ids,
                     end_ids=end_ids,
                     encoded_tags=encoded_tags,
                     text=text)
        for input_ids, attention_mask, token_type_ids, start_ids, end_ids,
        encoded_tags, input_len, token_offsets, text in zip(
            all_input_ids, all_attention_mask, all_token_type_ids,
            all_start_ids, all_end_ids, all_encoded_tags, all_input_lens,
            all_token_offsets, all_texts)
    ]

    return all_features


class PointerSequenceTagger(BaseTagger):
    def __init__(self, tokenizer, label2id):
        super(PointerSequenceTagger, self).__init__(tokenizer, label2id)

    def encode(self, examples, max_seq_length, **kwarg):
        #  tokenizer = kwarg.get('tokenizer', self.tokenizer)
        #  label2id = kwarg.get('label2id', self.label2id)
        tokenizer = self.tokenizer
        label2id = self.label2id
        return encode_examples(examples, tokenizer, label2id, max_seq_length)



# -*- coding: utf-8 -*-
# File: paste.py


import numpy as np
from abc import abstractmethod

from .base import ImageAugmentor
from .transform import TransformFactory

__all__ = ['CenterPaste', 'BackgroundFiller', 'ConstantBackgroundFiller',
           'RandomPaste']


class BackgroundFiller(object):
    """ Base class for all BackgroundFiller"""

    def fill(self, background_shape, img):
        """
        Return a proper background image of background_shape, given img.

        Args:
            background_shape (tuple): a shape (h, w)
            img: an image
        Returns:
            a background image
        """
        background_shape = tuple(background_shape)
        return self._fill(background_shape, img)

    @abstractmethod
    def _fill(self, background_shape, img):
        pass


class ConstantBackgroundFiller(BackgroundFiller):
    """ Fill the background by a constant """

    def __init__(self, value):
        """
        Args:
            value (float): the value to fill the background.
        """
        self.value = value

    def _fill(self, background_shape, img):
        assert img.ndim in [3, 2]
        if img.ndim == 3:
            return_shape = background_shape + (img.shape[2],)
        else:
            return_shape = background_shape
        return np.zeros(return_shape, dtype=img.dtype) + self.value


# NOTE:
# apply_coords should be implemeted in paste transform, but not yet done


class CenterPaste(ImageAugmentor):
    """
    Paste the image onto the center of a background canvas.
    """

    def __init__(self, background_shape, background_filler=None):
        """
        Args:
            background_shape (tuple): shape of the background canvas.
            background_filler (BackgroundFiller): How to fill the background. Defaults to zero-filler.
        """
        if background_filler is None:
            background_filler = ConstantBackgroundFiller(0)

        self._init(locals())

    def get_transform(self, _):
        return TransformFactory(name=str(self), apply_image=lambda img: self._impl(img))

    def _impl(self, img):
        img_shape = img.shape[:2]
        assert self.background_shape[0] >= img_shape[0] and self.background_shape[1] >= img_shape[1]

        background = self.background_filler.fill(
            self.background_shape, img)
        y0 = int((self.background_shape[0] - img_shape[0]) * 0.5)
        x0 = int((self.background_shape[1] - img_shape[1]) * 0.5)
        background[y0:y0 + img_shape[0], x0:x0 + img_shape[1]] = img
        return background


class RandomPaste(CenterPaste):
    """
    Randomly paste the image onto a background canvas.
    """

    def get_transform(self, img):
        img_shape = img.shape[:2]
        assert self.background_shape[0] > img_shape[0] and self.background_shape[1] > img_shape[1]

        y0 = self._rand_range(self.background_shape[0] - img_shape[0])
        x0 = self._rand_range(self.background_shape[1] - img_shape[1])
        l = int(x0), int(y0)
        return TransformFactory(name=str(self), apply_image=lambda img: self._impl(img, l))

    def _impl(self, img, loc):
        x0, y0 = loc
        img_shape = img.shape[:2]
        background = self.background_filler.fill(
            self.background_shape, img)
        background[y0:y0 + img_shape[0], x0:x0 + img_shape[1]] = img
        return background



# -*- coding: utf-8 -*-
# File: deform.py


import numpy as np

from ...utils import logger
from .base import ImageAugmentor
from .transform import TransformFactory

__all__ = []

# Code was temporarily kept here for a future reference in case someone needs it
# But it was already deprecated,
# because this augmentation is not a general one that people will often find helpful.


class GaussianMap(object):
    """ Generate Gaussian weighted deformation map"""
    # TODO really needs speedup

    def __init__(self, image_shape, sigma=0.5):
        assert len(image_shape) == 2
        self.shape = image_shape
        self.sigma = sigma

    def get_gaussian_weight(self, anchor):
        """
        Args:
            anchor: coordinate of the center
        """
        ret = np.zeros(self.shape, dtype='float32')

        y, x = np.mgrid[:self.shape[0], :self.shape[1]]
        y = y.astype('float32') / ret.shape[0] - anchor[0]
        x = x.astype('float32') / ret.shape[1] - anchor[1]
        g = np.exp(-(x**2 + y ** 2) / self.sigma)
        # cv2.imshow(" ", g)
        # cv2.waitKey()
        return g


def np_sample(img, coords):
    # a numpy implementation of ImageSample layer
    coords = np.maximum(coords, 0)
    coords = np.minimum(coords, np.array([img.shape[0] - 1, img.shape[1] - 1]))

    lcoor = np.floor(coords).astype('int32')
    ucoor = lcoor + 1
    ucoor = np.minimum(ucoor, np.array([img.shape[0] - 1, img.shape[1] - 1]))
    diff = coords - lcoor
    neg_diff = 1.0 - diff

    lcoory, lcoorx = np.split(lcoor, 2, axis=2)
    ucoory, ucoorx = np.split(ucoor, 2, axis=2)
    diff = np.repeat(diff, 3, 2).reshape((diff.shape[0], diff.shape[1], 2, 3))
    neg_diff = np.repeat(neg_diff, 3, 2).reshape((diff.shape[0], diff.shape[1], 2, 3))
    diffy, diffx = np.split(diff, 2, axis=2)
    ndiffy, ndiffx = np.split(neg_diff, 2, axis=2)

    ret = img[lcoory, lcoorx, :] * ndiffx * ndiffy + \
        img[ucoory, ucoorx, :] * diffx * diffy + \
        img[lcoory, ucoorx, :] * ndiffy * diffx + \
        img[ucoory, lcoorx, :] * diffy * ndiffx
    return ret[:, :, 0, :]


class GaussianDeform(ImageAugmentor):
    """
    Some kind of slow deformation I made up. Don't count on it.
    """

    # TODO input/output with different shape

    def __init__(self, anchors, shape, sigma=0.5, randrange=None):
        """
        Args:
            anchors (list): list of center coordinates in range [0,1].
            shape(list or tuple): image shape in [h, w].
            sigma (float): sigma for Gaussian weight
            randrange (int): offset range. Defaults to shape[0] / 8
        """
        logger.warn("GaussianDeform is slow. Consider using it with 4 or more prefetching processes.")
        super(GaussianDeform, self).__init__()
        self.anchors = anchors
        self.K = len(self.anchors)
        self.shape = shape
        self.grid = np.mgrid[0:self.shape[0], 0:self.shape[1]].transpose(1, 2, 0)
        self.grid = self.grid.astype('float32')  # HxWx2

        gm = GaussianMap(self.shape, sigma=sigma)
        self.gws = np.array([gm.get_gaussian_weight(ank)
                             for ank in self.anchors], dtype='float32')  # KxHxW
        self.gws = self.gws.transpose(1, 2, 0)  # HxWxK
        if randrange is None:
            self.randrange = self.shape[0] / 8
        else:
            self.randrange = randrange
        self.sigma = sigma

    def get_transform(self, img):
        v = self.rng.rand(self.K, 2).astype('float32') - 0.5
        v = v * 2 * self.randrange
        return TransformFactory(name=str(self), apply_image=lambda img: self._augment(img, v))

    def _augment(self, img, v):
        grid = self.grid + np.dot(self.gws, v)
        return np_sample(img, grid)



# -*- coding: utf-8 -*-
# File: parallel.py

import atexit
import pickle
import errno
import traceback
import itertools
import multiprocessing as mp
import os
import sys
import uuid
import weakref
from contextlib import contextmanager
import zmq
from six.moves import queue, range

from ..utils import logger
from ..utils.concurrency import (
    StoppableThread, enable_death_signal, ensure_proc_terminate, start_proc_mask_signal)
from ..utils.serialize import dumps_once as dumps, loads_once as loads
from .base import DataFlow, DataFlowReentrantGuard, DataFlowTerminated, ProxyDataFlow

__all__ = ['PrefetchData', 'MultiProcessPrefetchData',
           'MultiProcessRunner', 'MultiProcessRunnerZMQ', 'MultiThreadRunner',
           'PrefetchDataZMQ', 'MultiThreadPrefetchData']


# from https://github.com/pytorch/pytorch/blob/master/torch/utils/data/_utils/__init__.py
class _ExceptionWrapper:
    MAGIC = b"EXC_MAGIC"
    """Wraps an exception plus traceback to communicate across threads"""
    def __init__(self, exc_info):
        # It is important that we don't store exc_info, see
        # NOTE [ Python Traceback Reference Cycle Problem ]
        self.exc_type = exc_info[0]
        self.exc_msg = "".join(traceback.format_exception(*exc_info))

    def pack(self):
        return self.MAGIC + pickle.dumps(self)

    @staticmethod
    def unpack(dp):
        if isinstance(dp, bytes) and dp.startswith(_ExceptionWrapper.MAGIC):
            return pickle.loads(dp[len(_ExceptionWrapper.MAGIC):])


def _repeat_iter(get_itr):
    while True:
        yield from get_itr()


def _bind_guard(sock, name):
    try:
        sock.bind(name)
    except zmq.ZMQError:
        logger.error(
            "ZMQError in socket.bind('{}'). Perhaps you're \
using pipes on a non-local file system. See documentation of MultiProcessRunnerZMQ \
for more information.".format(name))
        raise


def _get_pipe_name(name):
    if sys.platform.startswith('linux'):
        # linux supports abstract sockets: http://api.zeromq.org/4-1:zmq-ipc
        pipename = "ipc://@{}-pipe-{}".format(name, str(uuid.uuid1())[:8])
        pipedir = os.environ.get('TENSORPACK_PIPEDIR', None)
        if pipedir is not None:
            logger.warn("TENSORPACK_PIPEDIR is not used on Linux any more! Abstract sockets will be used.")
    else:
        pipedir = os.environ.get('TENSORPACK_PIPEDIR', None)
        if pipedir is not None:
            logger.info("ZMQ uses TENSORPACK_PIPEDIR={}".format(pipedir))
        else:
            pipedir = '.'
        assert os.path.isdir(pipedir), pipedir
        filename = '{}/{}-pipe-{}'.format(pipedir.rstrip('/'), name, str(uuid.uuid1())[:6])
        assert not os.path.exists(filename), "Pipe {} exists! You may be unlucky.".format(filename)
        pipename = "ipc://{}".format(filename)
    return pipename


def del_weakref(x):
    o = x()
    if o is not None:
        o.__del__()


@contextmanager
def _zmq_catch_error(name):
    try:
        yield
    except zmq.ContextTerminated:
        logger.info("[{}] Context terminated.".format(name))
        raise DataFlowTerminated()
    except zmq.ZMQError as e:
        if e.errno == errno.ENOTSOCK:       # socket closed
            logger.info("[{}] Socket closed.".format(name))
            raise DataFlowTerminated()
        else:
            raise
    except Exception:
        raise


class _MultiProcessZMQDataFlow(DataFlow):
    def __init__(self):
        assert os.name != 'nt', "ZMQ IPC doesn't support windows!"
        self._reset_done = False
        self._procs = []

    def reset_state(self):
        """
        All forked dataflows should only be reset **once and only once** in spawned processes.
        Subclasses should call this method with super.
        """
        assert not self._reset_done, "reset_state() was called twice! This violates the API of DataFlow!"
        self._reset_done = True

        # __del__ not guaranteed to get called at exit
        atexit.register(del_weakref, weakref.ref(self))

    def _start_processes(self):
        start_proc_mask_signal(self._procs)

    def __del__(self):
        try:
            if not self._reset_done:
                return
            if not self.context.closed:
                self.socket.close(0)
                self.context.destroy(0)
            for x in self._procs:
                x.terminate()
                x.join(5)
            print("{} successfully cleaned-up.".format(type(self).__name__))
        except Exception:
            pass


class MultiProcessRunner(ProxyDataFlow):
    """
    Running a DataFlow in >=1 processes using Python multiprocessing utilities.
    It will fork the process that calls :meth:`__init__`, collect datapoints from `ds` in each
    process by a Python :class:`multiprocessing.Queue`.

    Note:
        1. (Data integrity) An iterator cannot run faster automatically -- what's happening is
           that the process will be forked ``num_proc`` times.
           There will be ``num_proc`` dataflow running in parallel and **independently**.
           As a result, we have the following guarantee on the dataflow correctness:

           a. When ``num_proc=1``, this dataflow produces the same data as the
              given dataflow in the same order.
           b. When ``num_proc>1``, if each sample from the given dataflow is i.i.d.,
              then this dataflow produces the **same distribution** of data as the given dataflow.
              This implies that there will be duplication, reordering, etc.
              You probably only want to use it for training.

              For example, if your original dataflow contains no randomness and produces the same first datapoint,
              then after parallel prefetching, the datapoint will be produced ``num_proc`` times
              at the beginning.
              Even when your original dataflow is fully shuffled, you still need to be aware of the
              `Birthday Paradox <https://en.wikipedia.org/wiki/Birthday_problem>`_
              and know that you'll likely see duplicates.

           To utilize parallelism with more strict data integrity, you can use
           the parallel versions of :class:`MapData`: :class:`MultiThreadMapData`, :class:`MultiProcessMapData`.
        2. This has more serialization overhead than :class:`MultiProcessRunnerZMQ` when data is large.
        3. You can nest like this: ``MultiProcessRunnerZMQ(MultiProcessRunner(df, num_proc=a), num_proc=b)``.
           A total of ``a`` instances of ``df`` worker processes will be created.
        4. Fork happens in `__init__`. `reset_state()` is a no-op.
           DataFlow in the worker processes will be reset at the time of fork.
        5. This DataFlow does support windows. However, Windows requires more strict picklability on processes,
           which means that some code that's forkable on Linux may not be forkable on Windows. If that happens you'll
           need to re-organize some part of code that's not forkable.
    """

    class _Worker(mp.Process):
        def __init__(self, ds, queue, idx):
            super(MultiProcessRunner._Worker, self).__init__()
            self.ds = ds
            self.queue = queue
            self.idx = idx

        def run(self):
            enable_death_signal(_warn=self.idx == 0)
            # reset all ds so each process will produce different data
            self.ds.reset_state()
            while True:
                for dp in self.ds:
                    self.queue.put(dp)

    def __init__(self, ds, num_prefetch, num_proc):
        """
        Args:
            ds (DataFlow): input DataFlow.
            num_prefetch (int): size of the queue to hold prefetched datapoints.
                Required.
            num_proc (int): number of processes to use. Required.
        """
        # https://docs.python.org/3.6/library/multiprocessing.html?highlight=process#the-spawn-and-forkserver-start-methods
        if os.name == 'nt':
            logger.warn("MultiProcessRunner does support Windows. \
However, Windows requires more strict picklability on processes, which may \
lead of failure on some of the code.")
        super(MultiProcessRunner, self).__init__(ds)
        try:
            self._size = len(ds)
        except NotImplementedError:
            self._size = -1
        assert num_proc > 0, num_proc
        assert num_prefetch > 0, num_prefetch
        self.num_proc = num_proc
        self.num_prefetch = num_prefetch

        if num_proc > 1:
            logger.info("[MultiProcessRunner] Will fork a dataflow more than one times. "
                        "This assumes the datapoints are i.i.d.")

        self.queue = mp.Queue(self.num_prefetch)
        self.procs = [MultiProcessRunner._Worker(self.ds, self.queue, idx)
                      for idx in range(self.num_proc)]
        ensure_proc_terminate(self.procs)
        self._reset_done = False

    def __iter__(self):
        for k in itertools.count():
            if self._size > 0 and k >= self._size:
                break
            dp = self.queue.get()
            yield dp

    def reset_state(self):
        assert not self._reset_done, "reset_state() was called twice! This violates the API of DataFlow!"
        self._reset_done = True
        start_proc_mask_signal(self.procs)


class MultiProcessRunnerZMQ(_MultiProcessZMQDataFlow):
    """
    Run a DataFlow in >=1 processes, with ZeroMQ for communication.
    It will fork the calling process of :meth:`reset_state()`,
    and collect datapoints from the given dataflow in each process by ZeroMQ IPC pipe.
    This is typically faster than :class:`MultiProcessRunner`.

    Note:
        1. (Data integrity) An iterator cannot run faster automatically -- what's happening is
           that the process will be forked ``num_proc`` times.
           There will be ``num_proc`` dataflow running in parallel and **independently**.
           As a result, we have the following guarantee on the dataflow correctness:

           a. When ``num_proc=1``, this dataflow produces the same data as the
              given dataflow in the same order.
           b. When ``num_proc>1``, if each sample from the given dataflow is i.i.d.,
              then this dataflow produces the **same distribution** of data as the given dataflow.
              This implies that there will be duplication, reordering, etc.
              You probably only want to use it for training.

              For example, if your original dataflow contains no randomness and produces the same first datapoint,
              then after parallel prefetching, the datapoint will be produced ``num_proc`` times
              at the beginning.
              Even when your original dataflow is fully shuffled, you still need to be aware of the
              `Birthday Paradox <https://en.wikipedia.org/wiki/Birthday_problem>`_
              and know that you'll likely see duplicates.

           To utilize parallelism with more strict data integrity, you can use
           the parallel versions of :class:`MapData`: :class:`MultiThreadMapData`, :class:`MultiProcessMapData`.
        2. `reset_state()` of the given dataflow will be called **once and only once** in the worker processes.
        3. The fork of processes happened in this dataflow's `reset_state()` method.
           Please note that forking a TensorFlow GPU session may be unsafe.
           If you're managing this dataflow on your own,
           it's better to fork before creating the session.
        4. (Fork-safety) After the fork has happened, this dataflow becomes not fork-safe.
           i.e., if you fork an already reset instance of this dataflow,
           it won't be usable in the forked process. Therefore, do not nest two `MultiProcessRunnerZMQ`.
        5. (Thread-safety) ZMQ is not thread safe. Therefore, do not call :meth:`get_data` of the same dataflow in
           more than 1 threads.
        6. This dataflow does not support windows. Use `MultiProcessRunner` which works on windows.
        7. (For Mac only) A UNIX named pipe will be created in the current directory.
           However, certain non-local filesystem such as NFS/GlusterFS/AFS doesn't always support pipes.
           You can change the directory by ``export TENSORPACK_PIPEDIR=/other/dir``.
           In particular, you can use somewhere under '/tmp' which is usually local.

           Note that some non-local FS may appear to support pipes and code
           may appear to run but crash with bizarre error.
           Also note that ZMQ limits the maximum length of pipe path.
           If you hit the limit, you can set the directory to a softlink
           which points to a local directory.
    """

    class _Worker(mp.Process):
        def __init__(self, ds, conn_name, hwm, idx):
            super(MultiProcessRunnerZMQ._Worker, self).__init__()
            self.ds = ds
            self.conn_name = conn_name
            self.hwm = hwm
            self.idx = idx

        def run(self):
            enable_death_signal(_warn=self.idx == 0)
            self.ds.reset_state()
            itr = _repeat_iter(lambda: self.ds)

            context = zmq.Context()
            socket = context.socket(zmq.PUSH)
            socket.set_hwm(self.hwm)
            socket.connect(self.conn_name)
            try:
                while True:
                    try:
                        dp = next(itr)
                        socket.send(dumps(dp), copy=False)
                    except Exception:
                        dp = _ExceptionWrapper(sys.exc_info()).pack()
                        socket.send(dumps(dp), copy=False)
                        raise
            # sigint could still propagate here, e.g. when nested
            except KeyboardInterrupt:
                pass
            finally:
                socket.close(0)
                context.destroy(0)

    def __init__(self, ds, num_proc=1, hwm=50):
        """
        Args:
            ds (DataFlow): input DataFlow.
            num_proc (int): number of processes to use.
            hwm (int): the zmq "high-water mark" (queue size) for both sender and receiver.
        """
        super(MultiProcessRunnerZMQ, self).__init__()

        self.ds = ds
        self.num_proc = num_proc
        self._hwm = hwm

        if num_proc > 1:
            logger.info("[MultiProcessRunnerZMQ] Will fork a dataflow more than one times. "
                        "This assumes the datapoints are i.i.d.")
        try:
            self._size = ds.__len__()
        except NotImplementedError:
            self._size = -1

    def _recv(self):
        ret = loads(self.socket.recv(copy=False))
        exc = _ExceptionWrapper.unpack(ret)
        if exc is not None:
            logger.error("Exception '{}' in worker:".format(str(exc.exc_type)))
            raise exc.exc_type(exc.exc_msg)
        return ret

    def __len__(self):
        return self.ds.__len__()

    def __iter__(self):
        with self._guard, _zmq_catch_error('MultiProcessRunnerZMQ'):
            for k in itertools.count():
                if self._size > 0 and k >= self._size:
                    break
                yield self._recv()

    def reset_state(self):
        super(MultiProcessRunnerZMQ, self).reset_state()
        self._guard = DataFlowReentrantGuard()
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PULL)
        self.socket.set_hwm(self._hwm)
        pipename = _get_pipe_name('dataflow')
        _bind_guard(self.socket, pipename)

        self._procs = [MultiProcessRunnerZMQ._Worker(self.ds, pipename, self._hwm, idx)
                       for idx in range(self.num_proc)]
        self._start_processes()


class MultiThreadRunner(DataFlow):
    """
    Create multiple dataflow instances and run them each in one thread.
    Collect outputs from them with a queue.

    Note:
        1. (Data integrity) An iterator cannot run faster automatically -- what's happening is
           that each thread will create a dataflow iterator.
           There will be ``num_thread`` dataflow running in parallel and **independently**.
           As a result, we have the following guarantee on the dataflow correctness:

           a. When ``num_thread=1``, this dataflow produces the same data as the
              given dataflow in the same order.
           b. When ``num_thread>1``, if each sample from the given dataflow is i.i.d.,
              then this dataflow produces the **same distribution** of data as the given dataflow.
              This implies that there will be duplication, reordering, etc.
              You probably only want to use it for training.

              For example, if your original dataflow contains no randomness and produces the same first datapoint,
              then after parallel prefetching, the datapoint will be produced ``num_thread`` times
              at the beginning.
              Even when your original dataflow is fully shuffled, you still need to be aware of the
              `Birthday Paradox <https://en.wikipedia.org/wiki/Birthday_problem>`_
              and know that you'll likely see duplicates.

           To utilize parallelism with more strict data integrity, you can use
           the parallel versions of :class:`MapData`: :class:`MultiThreadMapData`, :class:`MultiProcessMapData`.
    """

    class _Worker(StoppableThread):
        def __init__(self, get_df, queue):
            super(MultiThreadRunner._Worker, self).__init__()
            self.df = get_df()
            assert isinstance(self.df, DataFlow), self.df
            self.queue = queue
            self.daemon = True

        def run(self):
            self.df.reset_state()
            try:
                while True:
                    for dp in self.df:
                        if self.stopped():
                            return
                        self.queue_put_stoppable(self.queue, dp)
            except Exception:
                if self.stopped():
                    pass        # skip duplicated error messages
                else:
                    raise
            finally:
                self.stop()

    def __init__(self, get_df, num_prefetch, num_thread):
        """
        Args:
            get_df ( -> DataFlow): a callable which returns a DataFlow.
                Each thread will call this function to get the DataFlow to use.
                Therefore do not return the same DataFlow object for each call,
                unless your dataflow is stateless.
            num_prefetch (int): size of the queue
            num_thread (int): number of threads
        """
        assert num_thread > 0, num_thread
        assert num_prefetch > 0, num_prefetch
        self.num_thread = num_thread
        self.queue = queue.Queue(maxsize=num_prefetch)
        self.threads = [
            MultiThreadRunner._Worker(get_df, self.queue)
            for _ in range(num_thread)]

        try:
            self._size = self.__len__()
        except NotImplementedError:
            self._size = -1

    def reset_state(self):
        for th in self.threads:
            th.df.reset_state()
            th.start()

    def __len__(self):
        return self.threads[0].df.__len__()

    def __iter__(self):
        for k in itertools.count():
            if self._size > 0 and k >= self._size:
                break
            yield self.queue.get()

    def __del__(self):
        for p in self.threads:
            if p.is_alive():
                p.stop()
                p.join()


class PlasmaPutData(ProxyDataFlow):
    """
    Put each data point to plasma shared memory object store, and yield the object id instead.

    Experimental.
    """
    def __init__(self, ds, socket="/tmp/plasma"):
        self._socket = socket
        super(PlasmaPutData, self).__init__(ds)

    def reset_state(self):
        super(PlasmaPutData, self).reset_state()
        self.client = plasma.connect(self._socket, "", 0)

    def __iter__(self):
        for dp in self.ds:
            oid = self.client.put(dp)
            yield [oid.binary()]


class PlasmaGetData(ProxyDataFlow):
    """
    Take plasma object id from a DataFlow, and retrieve it from plasma shared
    memory object store.

    Experimental.
    """
    def __init__(self, ds, socket="/tmp/plasma"):
        self._socket = socket
        super(PlasmaGetData, self).__init__(ds)

    def reset_state(self):
        super(PlasmaGetData, self).reset_state()
        self.client = plasma.connect(self._socket, "", 0)

    def __iter__(self):
        for dp in self.ds:
            oid = plasma.ObjectID(dp[0])
            dp = self.client.get(oid)
            yield dp


plasma = None
# These plasma code is only experimental
# try:
#     import pyarrow.plasma as plasma
# except ImportError:
#     from ..utils.develop import create_dummy_class
#     PlasmaPutData = create_dummy_class('PlasmaPutData', 'pyarrow')   # noqa
#     PlasmaGetData = create_dummy_class('PlasmaGetData', 'pyarrow')   # noqa


# The old inappropriate names:
PrefetchData = MultiProcessRunner
MultiProcessPrefetchData = MultiProcessRunner
PrefetchDataZMQ = MultiProcessRunnerZMQ
MultiThreadPrefetchData = MultiThreadRunner

if __name__ == '__main__':
    import time
    from .raw import DataFromGenerator
    from .common import FixedSizeData
    x = DataFromGenerator(itertools.count())
    x = FixedSizeData(x, 100)
    x = MultiProcessRunnerZMQ(x, 2)
    x.reset_state()
    for idx, dp in enumerate(x):
        print(dp)
        time.sleep(0.1)
